/*
 * Copyright (c) 2006-2007, XenSource Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */


/*
 * Declarations of the abstract types used during demarshalling of enum
 * xen_on_crash_behaviour.  Internal to this library -- do not use from
 * outside.
 */


#ifndef XEN_ON_CRASH_BEHAVIOUR_INTERNAL_H
#define XEN_ON_CRASH_BEHAVIOUR_INTERNAL_H


#include "xen_internal.h"


extern const abstract_type xen_on_crash_behaviour_abstract_type_;
extern const abstract_type xen_on_crash_behaviour_set_abstract_type_;


#endif
